#!/usr/bin/env python3

##
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command line utility for reading and writing Avro files."""
import argparse
import csv
import functools
import itertools
import json
import sys
import warnings
from pathlib import Path
from typing import (
    IO,
    Any,
    AnyStr,
    Callable,
    Collection,
    Dict,
    Generator,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
)

import avro
import avro.datafile
import avro.errors
import avro.io
import avro.schema


def print_json(row: object) -> None:
    print(json.dumps(row))


def print_json_pretty(row: object) -> None:
    """Pretty print JSON"""
    # Need to work around https://bugs.python.org/issue16333
    # where json.dumps leaves trailing spaces.
    result = json.dumps(row, sort_keys=True, indent=4).replace(" \n", "\n")
    print(result)


_write_row = csv.writer(sys.stdout).writerow


def print_csv(row: Mapping[str, object]) -> None:
    # We sort the keys to the fields will be in the same place
    # FIXME: Do we want to do it in schema order?
    _write_row(row[key] for key in sorted(row))


def select_printer(format_: str) -> Callable[[Mapping[str, object]], None]:
    return {"json": print_json, "json-pretty": print_json_pretty, "csv": print_csv}[format_]


def record_match(expr: Any, record: Any) -> Callable[[Mapping[str, object]], bool]:
    warnings.warn(avro.errors.AvroWarning("There is no way to safely type check this."))
    return cast(Callable[[Mapping[str, object]], bool], eval(expr, None, {"r": record}))


def _passthrough_keys_filter(obj: Mapping[str, object]) -> Dict[str, object]:
    return {**obj}


def field_selector(fields: Collection[str]) -> Callable[[Mapping[str, object]], Dict[str, object]]:
    if fields:

        def keys_filter(obj: Mapping[str, object]) -> Dict[str, object]:
            return {k: obj[k] for k in obj.keys() & set(fields)}

        return keys_filter
    return _passthrough_keys_filter


def print_avro(
    data_file_reader: avro.datafile.DataFileReader,
    header: bool,
    format_: str,
    filter_: str,
    skip: int = 0,
    count: int = 0,
    fields: Collection[str] = "",
) -> None:
    if header and format_ != "csv":
        raise avro.errors.UsageError("--header applies only to CSV format")
    predicate = functools.partial(record_match, filter_) if filter_ else None
    avro_filtered = filter(predicate, data_file_reader)
    avro_slice = itertools.islice(avro_filtered, skip, (skip + count) if count else None)
    fs = field_selector(fields)
    avro_fields = (fs(cast(Mapping[str, object], r)) for r in avro_slice)
    avro_enum = enumerate(avro_fields)
    printer = select_printer(format_)
    for i, record in avro_enum:
        if header and i == 0:
            _write_row(sorted(record))
        printer(record)


def print_schema(data_file_reader: avro.datafile.DataFileReader) -> None:
    print(json.dumps(json.loads(data_file_reader.schema), indent=4))


def cat(
    files: Sequence[IO[AnyStr]], print_schema_: bool, header: bool, format_: str, filter_: str, skip: int, count: int, fields: Collection[str]
) -> int:
    datum_reader = avro.io.DatumReader()
    for file_ in files:
        with avro.datafile.DataFileReader(file_, datum_reader) as data_file_reader:
            if print_schema_:
                print_schema(data_file_reader)
                continue
            print_avro(data_file_reader, header, format_, filter_, skip, count, fields)
    return 0


def iter_json(info: Iterable[AnyStr], _: Any) -> Generator[object, None, None]:
    for i in info:
        row = (i if isinstance(i, str) else cast(bytes, i).decode()).strip()
        if row:
            yield json.loads(row)


def convert(value: str, field: avro.schema.Field) -> Union[int, float, str, bytes, bool, None]:
    type_ = field.type.type
    if type_ in ("int", "long"):
        return int(value)
    if type_ in ("float", "double"):
        return float(value)
    if type_ == "string":
        return value
    if type_ == "bytes":
        return value.encode()
    if type_ == "boolean":
        return value.lower() in ("1", "t", "true")
    if type_ == "null":
        return None
    if type_ == "union":
        return convert_union(value, field)
    raise avro.errors.UsageError("No valid conversion type")


def convert_union(value: str, field: avro.schema.Field) -> Union[int, float, str, bytes, bool, None]:
    for name in (s.name for s in field.type.schemas):
        try:
            return convert(value, name)
        except ValueError:
            continue
    raise avro.errors.UsageError("Exhausted Union Schema without finding a match")


def iter_csv(info: IO[AnyStr], schema: avro.schema.RecordSchema) -> Generator[Dict[str, object], None, None]:
    header = [field.name for field in schema.fields]
    for row in csv.reader(getattr(i, "decode", lambda: i)() for i in info):
        values = [convert(v, f) for v, f in zip(row, schema.fields)]
        yield dict(zip(header, values))


def guess_input_type(io_: IO[AnyStr]) -> str:
    ext = Path(io_.name).suffixes[0].lower()
    extensions = {".json": "json", ".js": "json", ".csv": "csv"}
    try:
        return extensions[ext]
    except KeyError as e:
        raise avro.errors.UsageError("Can't guess input file type (not .json or .csv)") from e


def write(schema: avro.schema.RecordSchema, output: IO[AnyStr], input_files: Sequence[IO[AnyStr]], input_type: Optional[str] = None) -> int:
    input_type = input_type or guess_input_type(cast(IO[AnyStr], next(iter(input_files), sys.stdin)))
    iter_records = {"json": iter_json, "csv": iter_csv}[input_type]
    with avro.datafile.DataFileWriter(output, avro.io.DatumWriter(), schema) as writer:
        for file_ in input_files:
            for record in iter_records(file_, schema):
                writer.append(record)
    return 0


def csv_arg(arg: Optional[str]) -> Optional[Tuple[str, ...]]:
    return tuple(f.strip() for f in arg.split(",") if f.strip()) if (arg and arg.strip()) else None


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Display/write for Avro files")
    parser.add_argument("--version", action="version", version=avro.__version__)
    subparsers = parser.add_subparsers(required=True, dest="command") if sys.version_info >= (3, 7) else parser.add_subparsers(dest="command")
    subparser_cat = subparsers.add_parser("cat")
    subparser_cat.add_argument("--version", action="version", version=avro.__version__)
    subparser_cat.add_argument(
        "--count",
        "-n",
        default=None,
        type=int,
        help="number of records to print",
    )
    subparser_cat.add_argument(
        "--skip",
        "-s",
        type=int,
        default=0,
        help="number of records to skip",
    )
    subparser_cat.add_argument(
        "--format",
        "-f",
        default="json",
        choices=("json", "csv", "json-pretty"),
        help="record format",
    )
    subparser_cat.add_argument(
        "--header",
        "-H",
        default=False,
        action="store_true",
        help="print CSV header",
    )
    subparser_cat.add_argument(
        "--filter",
        "-F",
        default=None,
        help="filter records (e.g. r['age']>1)",
    )
    subparser_cat.add_argument(
        "--print-schema",
        "-p",
        action="store_true",
        default=False,
        help="print schema",
    )
    subparser_cat.add_argument("--fields", type=csv_arg, default=None, help="fields to show, comma separated (show all by default)")
    subparser_cat.add_argument("files", nargs="*", type=argparse.FileType("rb"), help="Files to read", default=(sys.stdin,))
    subparser_write = subparsers.add_parser("write")
    subparser_write.add_argument("--version", action="version", version=avro.__version__)
    subparser_write.add_argument("--schema", "-s", type=avro.schema.from_path, required=True, help="schema JSON file (required)")
    subparser_write.add_argument(
        "--input-type",
        "-f",
        choices=("json", "csv"),
        default=None,
        help="input file(s) type (json or csv)",
    )
    subparser_write.add_argument("--output", "-o", help="output file", type=argparse.FileType("wb"), default=sys.stdout)
    subparser_write.add_argument("input_files", nargs="*", type=argparse.FileType("r"), help="Files to read", default=(sys.stdin,))
    return parser.parse_args()


def main() -> int:
    args = _parse_args()
    if args.command == "cat":
        return cat(args.files, args.print_schema, args.header, args.format, args.filter, args.skip, args.count, args.fields)
    if args.command == "write":
        return write(args.schema, args.output, args.input_files, args.input_type)
    raise avro.errors.UsageError(f"Unknown command - {args.command}")


if __name__ == "__main__":
    raise SystemExit(main())



#!/usr/bin/env python3

##
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import unittest

from avro.compatibility import (
    ReaderWriterCompatibilityChecker,
    SchemaCompatibilityType,
    SchemaType,
)
from avro.schema import (
    ArraySchema,
    MapSchema,
    Names,
    PrimitiveSchema,
    Schema,
    UnionSchema,
    parse,
)

BOOLEAN_SCHEMA = PrimitiveSchema(SchemaType.BOOLEAN)
NULL_SCHEMA = PrimitiveSchema(SchemaType.NULL)
INT_SCHEMA = PrimitiveSchema(SchemaType.INT)
LONG_SCHEMA = PrimitiveSchema(SchemaType.LONG)
STRING_SCHEMA = PrimitiveSchema(SchemaType.STRING)
BYTES_SCHEMA = PrimitiveSchema(SchemaType.BYTES)
FLOAT_SCHEMA = PrimitiveSchema(SchemaType.FLOAT)
DOUBLE_SCHEMA = PrimitiveSchema(SchemaType.DOUBLE)
INT_ARRAY_SCHEMA = ArraySchema(SchemaType.INT, names=Names())
LONG_ARRAY_SCHEMA = ArraySchema(SchemaType.LONG, names=Names())
STRING_ARRAY_SCHEMA = ArraySchema(SchemaType.STRING, names=Names())
INT_MAP_SCHEMA = MapSchema(SchemaType.INT, names=Names())
LONG_MAP_SCHEMA = MapSchema(SchemaType.LONG, names=Names())
STRING_MAP_SCHEMA = MapSchema(SchemaType.STRING, names=Names())
ENUM1_AB_SCHEMA = parse(json.dumps({"type": SchemaType.ENUM, "name": "Enum1", "symbols": ["A", "B"]}))
ENUM1_ABC_SCHEMA = parse(json.dumps({"type": SchemaType.ENUM, "name": "Enum1", "symbols": ["A", "B", "C"]}))
ENUM1_BC_SCHEMA = parse(json.dumps({"type": SchemaType.ENUM, "name": "Enum1", "symbols": ["B", "C"]}))
ENUM2_AB_SCHEMA = parse(json.dumps({"type": SchemaType.ENUM, "name": "Enum2", "symbols": ["A", "B"]}))
ENUM_ABC_ENUM_DEFAULT_A_SCHEMA = parse(json.dumps({"type": "enum", "name": "Enum", "symbols": ["A", "B", "C"], "default": "A"}))
ENUM_AB_ENUM_DEFAULT_A_SCHEMA = parse(json.dumps({"type": SchemaType.ENUM, "name": "Enum", "symbols": ["A", "B"], "default": "A"}))
ENUM_ABC_ENUM_DEFAULT_A_RECORD = parse(
    json.dumps(
        {
            "type": SchemaType.RECORD,
            "name": "Record",
            "fields": [
                {
                    "name": "Field",
                    "type": {
                        "type": SchemaType.ENUM,
                        "name": "Enum",
                        "symbols": ["A", "B", "C"],
                        "default": "A",
                    },
                }
            ],
        }
    )
)
ENUM_AB_ENUM_DEFAULT_A_RECORD = parse(
    json.dumps(
        {
            "type": SchemaType.RECORD,
            "name": "Record",
            "fields": [
                {
                    "name": "Field",
                    "type": {
                        "type": SchemaType.ENUM,
                        "name": "Enum",
                        "symbols": ["A", "B"],
                        "default": "A",
                    },
                }
            ],
        }
    )
)
ENUM_ABC_FIELD_DEFAULT_B_ENUM_DEFAULT_A_RECORD = parse(
    json.dumps(
        {
            "type": SchemaType.RECORD,
            "name": "Record",
            "fields": [
                {
                    "name": "Field",
                    "type": {
                        "type": SchemaType.ENUM,
                        "name": "Enum",
                        "symbols": ["A", "B", "C"],
                        "default": "A",
                    },
                    "default": "B",
                }
            ],
        }
    )
)
ENUM_AB_FIELD_DEFAULT_A_ENUM_DEFAULT_B_RECORD = parse(
    json.dumps(
        {
            "type": SchemaType.RECORD,
            "name": "Record",
            "fields": [
                {
                    "name": "Field",
                    "type": {
                        "type": SchemaType.ENUM,
                        "name": "Enum",
                        "symbols": ["A", "B"],
                        "default": "B",
                    },
                    "default": "A",
                }
            ],
        }
    )
)
EMPTY_UNION_SCHEMA = UnionSchema([], names=Names())
NULL_UNION_SCHEMA = UnionSchema([SchemaType.NULL], names=Names())
INT_UNION_SCHEMA = UnionSchema([SchemaType.INT], names=Names())
LONG_UNION_SCHEMA = UnionSchema([SchemaType.LONG], names=Names())
FLOAT_UNION_SCHEMA = UnionSchema([SchemaType.FLOAT], names=Names())
DOUBLE_UNION_SCHEMA = UnionSchema([SchemaType.DOUBLE], names=Names())
STRING_UNION_SCHEMA = UnionSchema([SchemaType.STRING], names=Names())
BYTES_UNION_SCHEMA = UnionSchema([SchemaType.BYTES], names=Names())
INT_STRING_UNION_SCHEMA = UnionSchema([SchemaType.INT, SchemaType.STRING], names=Names())
STRING_INT_UNION_SCHEMA = UnionSchema([SchemaType.STRING, SchemaType.INT], names=Names())
INT_FLOAT_UNION_SCHEMA = UnionSchema([SchemaType.INT, SchemaType.FLOAT], names=Names())
INT_LONG_UNION_SCHEMA = UnionSchema([SchemaType.INT, SchemaType.LONG], names=Names())
INT_LONG_FLOAT_DOUBLE_UNION_SCHEMA = UnionSchema(
    [SchemaType.INT, SchemaType.LONG, SchemaType.FLOAT, SchemaType.DOUBLE],
    names=Names(),
)
NULL_INT_ARRAY_UNION_SCHEMA = UnionSchema(
    [{"type": SchemaType.NULL}, {"type": SchemaType.ARRAY, "items": SchemaType.INT}],
    names=Names(),
)
NULL_INT_MAP_UNION_SCHEMA = UnionSchema(
    [{"type": SchemaType.NULL}, {"type": SchemaType.MAP, "values": SchemaType.INT}],
    names=Names(),
)
EMPTY_RECORD1 = parse(json.dumps({"type": SchemaType.RECORD, "name": "Record1", "fields": []}))
EMPTY_RECORD2 = parse(json.dumps({"type": SchemaType.RECORD, "name": "Record2", "fields": []}))
A_INT_RECORD1 = parse(
    json.dumps(
        {
            "type": SchemaType.RECORD,
            "name": "Record1",
            "fields": [{"name": "a", "type": SchemaType.INT}],
        }
    )
)
A_LONG_RECORD1 = parse(
    json.dumps(
        {
            "type": SchemaType.RECORD,
            "name": "Record1",
            "fields": [{"name": "a", "type": SchemaType.LONG}],
        }
    )
)
A_INT_B_INT_RECORD1 = parse(
    json.dumps(
        {
            "type": SchemaType.RECORD,
            "name": "Record1",
            "fields": [
                {"name": "a", "type": SchemaType.INT},
                {"name": "b", "type": SchemaType.INT},
            ],
        }
    )
)
A_DINT_RECORD1 = parse(
    json.dumps(
        {
            "type": SchemaType.RECORD,
            "name": "Record1",
            "fields": [{"name": "a", "type": SchemaType.INT, "default": 0}],
        }
    )
)
A_INT_B_DINT_RECORD1 = parse(
    json.dumps(
        {
            "type": SchemaType.RECORD,
            "name": "Record1",
            "fields": [
                {"name": "a", "type": SchemaType.INT},
                {"name": "b", "type": SchemaType.INT, "default": 0},
            ],
        }
    )
)
A_DINT_B_DINT_RECORD1 = parse(
    json.dumps(
        {
            "type": SchemaType.RECORD,
            "name": "Record1",
            "fields": [
                {"name": "a", "type": SchemaType.INT, "default": 0},
                {"name": "b", "type": SchemaType.INT, "default": 0},
            ],
        }
    )
)
A_DINT_B_DFIXED_4_BYTES_RECORD1 = parse(
    json.dumps(
        {
            "type": SchemaType.RECORD,
            "name": "Record1",
            "fields": [
                {"name": "a", "type": SchemaType.INT, "default": 0},
                {
                    "name": "b",
                    "type": {"type": SchemaType.FIXED, "name": "Fixed", "size": 4},
                },
            ],
        }
    )
)
A_DINT_B_DFIXED_8_BYTES_RECORD1 = parse(
    json.dumps(
        {
            "type": SchemaType.RECORD,
            "name": "Record1",
            "fields": [
                {"name": "a", "type": SchemaType.INT, "default": 0},
                {
                    "name": "b",
                    "type": {"type": SchemaType.FIXED, "name": "Fixed", "size": 8},
                },
            ],
        }
    )
)
A_DINT_B_DINT_STRING_UNION_RECORD1 = parse(
    json.dumps(
        {
            "type": SchemaType.RECORD,
            "name": "Record1",
            "fields": [
                {"name": "a", "type": SchemaType.INT, "default": 0},
                {
                    "name": "b",
                    "type": [SchemaType.INT, SchemaType.STRING],
                    "default": 0,
                },
            ],
        }
    )
)
A_DINT_B_DINT_UNION_RECORD1 = parse(
    json.dumps(
        {
            "type": SchemaType.RECORD,
            "name": "Record1",
            "fields": [
                {"name": "a", "type": SchemaType.INT, "default": 0},
                {"name": "b", "type": [SchemaType.INT], "default": 0},
            ],
        }
    )
)
A_DINT_B_DENUM_1_RECORD1 = parse(
    json.dumps(
        {
            "type": SchemaType.RECORD,
            "name": "Record1",
            "fields": [
                {"name": "a", "type": SchemaType.INT, "default": 0},
                {
                    "name": "b",
                    "type": {
                        "type": SchemaType.ENUM,
                        "name": "Enum1",
                        "symbols": ["A", "B"],
                    },
                },
            ],
        }
    )
)
A_DINT_B_DENUM_2_RECORD1 = parse(
    json.dumps(
        {
            "type": SchemaType.RECORD,
            "name": "Record1",
            "fields": [
                {"name": "a", "type": SchemaType.INT, "default": 0},
                {
                    "name": "b",
                    "type": {
                        "type": SchemaType.ENUM,
                        "name": "Enum2",
                        "symbols": ["A", "B"],
                    },
                },
            ],
        }
    )
)
FIXED_4_BYTES = parse(json.dumps({"type": SchemaType.FIXED, "name": "Fixed", "size": 4}))
FIXED_8_BYTES = parse(json.dumps({"type": SchemaType.FIXED, "name": "Fixed", "size": 8}))
NS_RECORD1 = parse(
    json.dumps(
        {
            "type": SchemaType.RECORD,
            "name": "Record1",
            "fields": [
                {
                    "name": "f1",
                    "type": [
                        SchemaType.NULL,
                        {
                            "type": SchemaType.ARRAY,
                            "items": {
                                "type": SchemaType.RECORD,
                                "name": "InnerRecord1",
                                "namespace": "ns1",
                                "fields": [{"name": "a", "type": SchemaType.INT}],
                            },
                        },
                    ],
                }
            ],
        }
    )
)
NS_RECORD2 = parse(
    json.dumps(
        {
            "type": SchemaType.RECORD,
            "name": "Record1",
            "fields": [
                {
                    "name": "f1",
                    "type": [
                        SchemaType.NULL,
                        {
                            "type": SchemaType.ARRAY,
                            "items": {
                                "type": SchemaType.RECORD,
                                "name": "InnerRecord1",
                                "namespace": "ns2",
                                "fields": [{"name": "a", "type": SchemaType.INT}],
                            },
                        },
                    ],
                }
            ],
        }
    )
)

UNION_INT_RECORD1 = UnionSchema(
    [
        {"type": SchemaType.INT},
        {
            "type": SchemaType.RECORD,
            "name": "Record1",
            "fields": [{"name": "field1", "type": SchemaType.INT}],
        },
    ]
)
UNION_INT_RECORD2 = UnionSchema(
    [
        {"type": SchemaType.INT},
        {
            "type": "record",
            "name": "Record2",
            "fields": [{"name": "field1", "type": SchemaType.INT}],
        },
    ]
)
UNION_INT_ENUM1_AB = UnionSchema([{"type": SchemaType.INT}, ENUM1_AB_SCHEMA.to_json()])
UNION_INT_FIXED_4_BYTES = UnionSchema([{"type": SchemaType.INT}, FIXED_4_BYTES.to_json()])
UNION_INT_BOOLEAN = UnionSchema([{"type": SchemaType.INT}, {"type": SchemaType.BOOLEAN}])
UNION_INT_ARRAY_INT = UnionSchema([{"type": SchemaType.INT}, INT_ARRAY_SCHEMA.to_json()])
UNION_INT_MAP_INT = UnionSchema([{"type": SchemaType.INT}, INT_MAP_SCHEMA.to_json()])
UNION_INT_NULL = UnionSchema([{"type": SchemaType.INT}, {"type": SchemaType.NULL}])
FIXED_4_ANOTHER_NAME = parse(json.dumps({"type": SchemaType.FIXED, "name": "AnotherName", "size": 4}))
RECORD1_WITH_ENUM_AB = parse(
    json.dumps(
        {
            "type": SchemaType.RECORD,
            "name": "Record1",
            "fields": [{"name": "field1", "type": ENUM1_AB_SCHEMA.to_json()}],
        }
    )
)
RECORD1_WITH_ENUM_ABC = parse(
    json.dumps(
        {
            "type": SchemaType.RECORD,
            "name": "Record1",
            "fields": [{"name": "field1", "type": ENUM1_ABC_SCHEMA.to_json()}],
        }
    )
)


class TestCompatibility(unittest.TestCase):
    def test_simple_schema_promotion(self):
        field_alias_reader = parse(
            json.dumps(
                {
                    "name": "foo",
                    "type": "record",
                    "fields": [{"type": "int", "name": "bar", "aliases": ["f1"]}],
                }
            )
        )
        record_alias_reader = parse(
            json.dumps(
                {
                    "name": "other",
                    "type": "record",
                    "fields": [{"type": "int", "name": "f1"}],
                    "aliases": ["foo"],
                }
            )
        )

        writer = parse(
            json.dumps(
                {
                    "name": "foo",
                    "type": "record",
                    "fields": [
                        {"type": "int", "name": "f1"},
                        {
                            "type": "string",
                            "name": "f2",
                        },
                    ],
                }
            )
        )
        # alias testing
        res = ReaderWriterCompatibilityChecker().get_compatibility(field_alias_reader, writer)
        self.assertIs(res.compatibility, SchemaCompatibilityType.compatible, res.locations)
        res = ReaderWriterCompatibilityChecker().get_compatibility(record_alias_reader, writer)
        self.assertIs(res.compatibility, SchemaCompatibilityType.compatible, res.locations)

    def test_schema_compatibility(self):
        # testValidateSchemaPairMissingField
        writer = parse(
            json.dumps(
                {
                    "type": SchemaType.RECORD,
                    "name": "Record",
                    "fields": [
                        {"name": "oldField1", "type": SchemaType.INT},
                        {"name": "oldField2", "type": SchemaType.STRING},
                    ],
                }
            )
        )
        reader = parse(
            json.dumps(
                {
                    "type": SchemaType.RECORD,
                    "name": "Record",
                    "fields": [{"name": "oldField1", "type": SchemaType.INT}],
                }
            )
        )
        self.assertTrue(self.are_compatible(reader, writer))
        # testValidateSchemaPairMissingSecondField
        reader = parse(
            json.dumps(
                {
                    "type": SchemaType.RECORD,
                    "name": "Record",
                    "fields": [{"name": "oldField2", "type": SchemaType.STRING}],
                }
            )
        )
        self.assertTrue(self.are_compatible(reader, writer))
        # testValidateSchemaPairAllFields
        reader = parse(
            json.dumps(
                {
                    "type": SchemaType.RECORD,
                    "name": "Record",
                    "fields": [
                        {"name": "oldField1", "type": SchemaType.INT},
                        {"name": "oldField2", "type": SchemaType.STRING},
                    ],
                }
            )
        )
        self.assertTrue(self.are_compatible(reader, writer))
        # testValidateSchemaNewFieldWithDefault
        reader = parse(
            json.dumps(
                {
                    "type": SchemaType.RECORD,
                    "name": "Record",
                    "fields": [
                        {"name": "oldField1", "type": SchemaType.INT},
                        {"name": "newField2", "type": SchemaType.INT, "default": 42},
                    ],
                }
            )
        )
        self.assertTrue(self.are_compatible(reader, writer))
        # testValidateSchemaNewField
        reader = parse(
            json.dumps(
                {
                    "type": SchemaType.RECORD,
                    "name": "Record",
                    "fields": [
                        {"name": "oldField1", "type": SchemaType.INT},
                        {"name": "newField2", "type": SchemaType.INT},
                    ],
                }
            )
        )
        self.assertFalse(self.are_compatible(reader, writer))
        # testValidateArrayWriterSchema
        writer = parse(json.dumps({"type": SchemaType.ARRAY, "items": {"type": SchemaType.STRING}}))
        reader = parse(json.dumps({"type": SchemaType.ARRAY, "items": {"type": SchemaType.STRING}}))
        self.assertTrue(self.are_compatible(reader, writer))
        reader = parse(json.dumps({"type": SchemaType.MAP, "values": {"type": SchemaType.STRING}}))
        self.assertFalse(self.are_compatible(reader, writer))
        # testValidatePrimitiveWriterSchema
        writer = parse(json.dumps({"type": SchemaType.STRING}))
        reader = parse(json.dumps({"type": SchemaType.STRING}))
        self.assertTrue(self.are_compatible(reader, writer))
        reader = parse(json.dumps({"type": SchemaType.INT}))
        self.assertFalse(self.are_compatible(reader, writer))
        # testUnionReaderWriterSubsetIncompatibility
        writer = parse(
            json.dumps(
                {
                    "name": "Record",
                    "type": "record",
                    "fields": [
                        {
                            "name": "f1",
                            "type": [
                                SchemaType.INT,
                                SchemaType.STRING,
                                SchemaType.LONG,
                            ],
                        }
                    ],
                }
            )
        )
        reader = parse(
            json.dumps(
                {
                    "name": "Record",
                    "type": SchemaType.RECORD,
                    "fields": [{"name": "f1", "type": [SchemaType.INT, SchemaType.STRING]}],
                }
            )
        )
        reader = reader.fields[0].type
        writer = writer.fields[0].type
        self.assertIsInstance(reader, UnionSchema)
        self.assertIsInstance(writer, UnionSchema)
        self.assertFalse(self.are_compatible(reader, writer))
        # testReaderWriterCompatibility
        compatible_reader_writer_test_cases = [
            (BOOLEAN_SCHEMA, BOOLEAN_SCHEMA),
            (INT_SCHEMA, INT_SCHEMA),
            (LONG_SCHEMA, INT_SCHEMA),
            (LONG_SCHEMA, LONG_SCHEMA),
            (FLOAT_SCHEMA, INT_SCHEMA),
            (FLOAT_SCHEMA, LONG_SCHEMA),
            (DOUBLE_SCHEMA, LONG_SCHEMA),
            (DOUBLE_SCHEMA, INT_SCHEMA),
            (DOUBLE_SCHEMA, FLOAT_SCHEMA),
            (STRING_SCHEMA, STRING_SCHEMA),
            (BYTES_SCHEMA, BYTES_SCHEMA),
            (STRING_SCHEMA, BYTES_SCHEMA),
            (BYTES_SCHEMA, STRING_SCHEMA),
            (INT_ARRAY_SCHEMA, INT_ARRAY_SCHEMA),
            (LONG_ARRAY_SCHEMA, INT_ARRAY_SCHEMA),
            (INT_MAP_SCHEMA, INT_MAP_SCHEMA),
            (LONG_MAP_SCHEMA, INT_MAP_SCHEMA),
            (ENUM1_AB_SCHEMA, ENUM1_AB_SCHEMA),
            (ENUM1_ABC_SCHEMA, ENUM1_AB_SCHEMA),
            # Union related pairs
            (EMPTY_UNION_SCHEMA, EMPTY_UNION_SCHEMA),
            (FLOAT_UNION_SCHEMA, EMPTY_UNION_SCHEMA),
            (FLOAT_UNION_SCHEMA, INT_UNION_SCHEMA),
            (FLOAT_UNION_SCHEMA, LONG_UNION_SCHEMA),
            (FLOAT_UNION_SCHEMA, INT_LONG_UNION_SCHEMA),
            (INT_UNION_SCHEMA, INT_UNION_SCHEMA),
            (INT_STRING_UNION_SCHEMA, STRING_INT_UNION_SCHEMA),
            (INT_UNION_SCHEMA, EMPTY_UNION_SCHEMA),
            (LONG_UNION_SCHEMA, EMPTY_UNION_SCHEMA),
            (LONG_UNION_SCHEMA, INT_UNION_SCHEMA),
            (FLOAT_UNION_SCHEMA, INT_UNION_SCHEMA),
            (DOUBLE_UNION_SCHEMA, INT_UNION_SCHEMA),
            (FLOAT_UNION_SCHEMA, LONG_UNION_SCHEMA),
            (DOUBLE_UNION_SCHEMA, LONG_UNION_SCHEMA),
            (FLOAT_UNION_SCHEMA, EMPTY_UNION_SCHEMA),
            (DOUBLE_UNION_SCHEMA, FLOAT_UNION_SCHEMA),
            (STRING_UNION_SCHEMA, EMPTY_UNION_SCHEMA),
            (STRING_UNION_SCHEMA, BYTES_UNION_SCHEMA),
            (BYTES_UNION_SCHEMA, EMPTY_UNION_SCHEMA),
            (BYTES_UNION_SCHEMA, STRING_UNION_SCHEMA),
            (DOUBLE_UNION_SCHEMA, INT_FLOAT_UNION_SCHEMA),
            # Readers capable of reading all branches of a union are compatible
            (FLOAT_SCHEMA, INT_FLOAT_UNION_SCHEMA),
            (LONG_SCHEMA, INT_LONG_UNION_SCHEMA),
            (DOUBLE_SCHEMA, INT_FLOAT_UNION_SCHEMA),
            (DOUBLE_SCHEMA, INT_LONG_FLOAT_DOUBLE_UNION_SCHEMA),
            # Special case of singleton unions:
            (FLOAT_SCHEMA, FLOAT_UNION_SCHEMA),
            (INT_UNION_SCHEMA, INT_SCHEMA),
            (INT_SCHEMA, INT_UNION_SCHEMA),
            # Fixed types
            (FIXED_4_BYTES, FIXED_4_BYTES),
            # Tests involving records:
            (EMPTY_RECORD1, EMPTY_RECORD1),
            (EMPTY_RECORD1, A_INT_RECORD1),
            (A_INT_RECORD1, A_INT_RECORD1),
            (A_DINT_RECORD1, A_INT_RECORD1),
            (A_DINT_RECORD1, A_DINT_RECORD1),
            (A_INT_RECORD1, A_DINT_RECORD1),
            (A_LONG_RECORD1, A_INT_RECORD1),
            (A_INT_RECORD1, A_INT_B_INT_RECORD1),
            (A_DINT_RECORD1, A_INT_B_INT_RECORD1),
            (A_INT_B_DINT_RECORD1, A_INT_RECORD1),
            (A_DINT_B_DINT_RECORD1, EMPTY_RECORD1),
            (A_DINT_B_DINT_RECORD1, A_INT_RECORD1),
            (A_INT_B_INT_RECORD1, A_DINT_B_DINT_RECORD1),
            (parse(json.dumps({"type": "null"})), parse(json.dumps({"type": "null"}))),
            (NULL_SCHEMA, NULL_SCHEMA),
            (ENUM_AB_ENUM_DEFAULT_A_RECORD, ENUM_ABC_ENUM_DEFAULT_A_RECORD),
            (
                ENUM_AB_FIELD_DEFAULT_A_ENUM_DEFAULT_B_RECORD,
                ENUM_ABC_FIELD_DEFAULT_B_ENUM_DEFAULT_A_RECORD,
            ),
            (NS_RECORD1, NS_RECORD2),
        ]

        for (reader, writer) in compatible_reader_writer_test_cases:
            self.assertTrue(self.are_compatible(reader, writer))

    def test_schema_compatibility_fixed_size_mismatch(self):
        incompatible_fixed_pairs = [
            (FIXED_4_BYTES, FIXED_8_BYTES, "expected: 8, found: 4", "/size"),
            (FIXED_8_BYTES, FIXED_4_BYTES, "expected: 4, found: 8", "/size"),
            (
                A_DINT_B_DFIXED_8_BYTES_RECORD1,
                A_DINT_B_DFIXED_4_BYTES_RECORD1,
                "expected: 4, found: 8",
                "/fields/1/type/size",
            ),
            (
                A_DINT_B_DFIXED_4_BYTES_RECORD1,
                A_DINT_B_DFIXED_8_BYTES_RECORD1,
                "expected: 8, found: 4",
                "/fields/1/type/size",
            ),
        ]
        for (reader, writer, message, location) in incompatible_fixed_pairs:
            result = ReaderWriterCompatibilityChecker().get_compatibility(reader, writer)
            self.assertIs(result.compatibility, SchemaCompatibilityType.incompatible)
            self.assertIn(
                location,
                result.locations,
                f"expected {location}, found {result}",
            )
            self.assertIn(
                message,
                result.messages,
                f"expected {location}, found {result}",
            )

    def test_schema_compatibility_missing_enum_symbols(self):
        incompatible_pairs = [
            # str(set) representation
            (ENUM1_AB_SCHEMA, ENUM1_ABC_SCHEMA, "{'C'}", "/symbols"),
            (ENUM1_BC_SCHEMA, ENUM1_ABC_SCHEMA, "{'A'}", "/symbols"),
            (
                RECORD1_WITH_ENUM_AB,
                RECORD1_WITH_ENUM_ABC,
                "{'C'}",
                "/fields/0/type/symbols",
            ),
        ]
        for (reader, writer, message, location) in incompatible_pairs:
            result = ReaderWriterCompatibilityChecker().get_compatibility(reader, writer)
            self.assertIs(result.compatibility, SchemaCompatibilityType.incompatible)
            self.assertIn(message, result.messages)
            self.assertIn(location, result.locations)

    def test_schema_compatibility_missing_union_branch(self):
        incompatible_pairs = [
            (
                INT_UNION_SCHEMA,
                INT_STRING_UNION_SCHEMA,
                {"reader union lacking writer type: STRING"},
                {"/1"},
            ),
            (
                STRING_UNION_SCHEMA,
                INT_STRING_UNION_SCHEMA,
                {"reader union lacking writer type: INT"},
                {"/0"},
            ),
            (
                INT_UNION_SCHEMA,
                UNION_INT_RECORD1,
                {"reader union lacking writer type: RECORD"},
                {"/1"},
            ),
            (
                INT_UNION_SCHEMA,
                UNION_INT_RECORD2,
                {"reader union lacking writer type: RECORD"},
                {"/1"},
            ),
            (
                UNION_INT_RECORD1,
                UNION_INT_RECORD2,
                {"reader union lacking writer type: RECORD"},
                {"/1"},
            ),
            (
                INT_UNION_SCHEMA,
                UNION_INT_ENUM1_AB,
                {"reader union lacking writer type: ENUM"},
                {"/1"},
            ),
            (
                INT_UNION_SCHEMA,
                UNION_INT_FIXED_4_BYTES,
                {"reader union lacking writer type: FIXED"},
                {"/1"},
            ),
            (
                INT_UNION_SCHEMA,
                UNION_INT_BOOLEAN,
                {"reader union lacking writer type: BOOLEAN"},
                {"/1"},
            ),
            (
                INT_UNION_SCHEMA,
                LONG_UNION_SCHEMA,
                {"reader union lacking writer type: LONG"},
                {"/0"},
            ),
            (
                INT_UNION_SCHEMA,
                FLOAT_UNION_SCHEMA,
                {"reader union lacking writer type: FLOAT"},
                {"/0"},
            ),
            (
                INT_UNION_SCHEMA,
                DOUBLE_UNION_SCHEMA,
                {"reader union lacking writer type: DOUBLE"},
                {"/0"},
            ),
            (
                INT_UNION_SCHEMA,
                BYTES_UNION_SCHEMA,
                {"reader union lacking writer type: BYTES"},
                {"/0"},
            ),
            (
                INT_UNION_SCHEMA,
                UNION_INT_ARRAY_INT,
                {"reader union lacking writer type: ARRAY"},
                {"/1"},
            ),
            (
                INT_UNION_SCHEMA,
                UNION_INT_MAP_INT,
                {"reader union lacking writer type: MAP"},
                {"/1"},
            ),
            (
                INT_UNION_SCHEMA,
                UNION_INT_NULL,
                {"reader union lacking writer type: NULL"},
                {"/1"},
            ),
            (
                INT_UNION_SCHEMA,
                INT_LONG_FLOAT_DOUBLE_UNION_SCHEMA,
                {
                    "reader union lacking writer type: LONG",
                    "reader union lacking writer type: FLOAT",
                    "reader union lacking writer type: DOUBLE",
                },
                {"/1", "/2", "/3"},
            ),
            (
                A_DINT_B_DINT_UNION_RECORD1,
                A_DINT_B_DINT_STRING_UNION_RECORD1,
                {"reader union lacking writer type: STRING"},
                {"/fields/1/type/1"},
            ),
        ]

        for (reader, writer, message, location) in incompatible_pairs:
            result = ReaderWriterCompatibilityChecker().get_compatibility(reader, writer)
            self.assertIs(result.compatibility, SchemaCompatibilityType.incompatible)
            self.assertEqual(result.messages, message)
            self.assertEqual(result.locations, location)

    def test_schema_compatibility_name_mismatch(self):
        incompatible_pairs = [
            (ENUM1_AB_SCHEMA, ENUM2_AB_SCHEMA, "expected: Enum2", "/name"),
            (EMPTY_RECORD2, EMPTY_RECORD1, "expected: Record1", "/name"),
            (FIXED_4_BYTES, FIXED_4_ANOTHER_NAME, "expected: AnotherName", "/name"),
            (
                A_DINT_B_DENUM_1_RECORD1,
                A_DINT_B_DENUM_2_RECORD1,
                "expected: Enum2",
                "/fields/1/type/name",
            ),
        ]

        for (reader, writer, message, location) in incompatible_pairs:
            result = ReaderWriterCompatibilityChecker().get_compatibility(reader, writer)
            self.assertIs(result.compatibility, SchemaCompatibilityType.incompatible)
            self.assertIn(message, result.messages)
            self.assertIn(location, result.locations)

    def test_schema_compatibility_reader_field_missing_default_value(self):
        incompatible_pairs = [
            (A_INT_RECORD1, EMPTY_RECORD1, "a", "/fields/0"),
            (A_INT_B_DINT_RECORD1, EMPTY_RECORD1, "a", "/fields/0"),
        ]
        for (reader, writer, message, location) in incompatible_pairs:
            result = ReaderWriterCompatibilityChecker().get_compatibility(reader, writer)
            self.assertIs(result.compatibility, SchemaCompatibilityType.incompatible)
            self.assertEqual(len(result.messages), 1)
            self.assertEqual(len(result.locations), 1)
            self.assertEqual(message, "".join(result.messages))
            self.assertEqual(location, "".join(result.locations))

    def test_schema_compatibility_type_mismatch(self):
        incompatible_pairs = [
            (
                NULL_SCHEMA,
                INT_SCHEMA,
                "reader type: null not compatible with writer type: int",
                "/",
            ),
            (
                NULL_SCHEMA,
                LONG_SCHEMA,
                "reader type: null not compatible with writer type: long",
                "/",
            ),
            (
                BOOLEAN_SCHEMA,
                INT_SCHEMA,
                "reader type: boolean not compatible with writer type: int",
                "/",
            ),
            (
                INT_SCHEMA,
                NULL_SCHEMA,
                "reader type: int not compatible with writer type: null",
                "/",
            ),
            (
                INT_SCHEMA,
                BOOLEAN_SCHEMA,
                "reader type: int not compatible with writer type: boolean",
                "/",
            ),
            (
                INT_SCHEMA,
                LONG_SCHEMA,
                "reader type: int not compatible with writer type: long",
                "/",
            ),
            (
                INT_SCHEMA,
                FLOAT_SCHEMA,
                "reader type: int not compatible with writer type: float",
                "/",
            ),
            (
                INT_SCHEMA,
                DOUBLE_SCHEMA,
                "reader type: int not compatible with writer type: double",
                "/",
            ),
            (
                LONG_SCHEMA,
                FLOAT_SCHEMA,
                "reader type: long not compatible with writer type: float",
                "/",
            ),
            (
                LONG_SCHEMA,
                DOUBLE_SCHEMA,
                "reader type: long not compatible with writer type: double",
                "/",
            ),
            (
                FLOAT_SCHEMA,
                DOUBLE_SCHEMA,
                "reader type: float not compatible with writer type: double",
                "/",
            ),
            (
                DOUBLE_SCHEMA,
                STRING_SCHEMA,
                "reader type: double not compatible with writer type: string",
                "/",
            ),
            (
                FIXED_4_BYTES,
                STRING_SCHEMA,
                "reader type: fixed not compatible with writer type: string",
                "/",
            ),
            (
                STRING_SCHEMA,
                BOOLEAN_SCHEMA,
                "reader type: string not compatible with writer type: boolean",
                "/",
            ),
            (
                STRING_SCHEMA,
                INT_SCHEMA,
                "reader type: string not compatible with writer type: int",
                "/",
            ),
            (
                BYTES_SCHEMA,
                NULL_SCHEMA,
                "reader type: bytes not compatible with writer type: null",
                "/",
            ),
            (
                BYTES_SCHEMA,
                INT_SCHEMA,
                "reader type: bytes not compatible with writer type: int",
                "/",
            ),
            (
                A_INT_RECORD1,
                INT_SCHEMA,
                "reader type: record not compatible with writer type: int",
                "/",
            ),
            (
                INT_ARRAY_SCHEMA,
                LONG_ARRAY_SCHEMA,
                "reader type: int not compatible with writer type: long",
                "/items",
            ),
            (
                INT_MAP_SCHEMA,
                INT_ARRAY_SCHEMA,
                "reader type: map not compatible with writer type: array",
                "/",
            ),
            (
                INT_ARRAY_SCHEMA,
                INT_MAP_SCHEMA,
                "reader type: array not compatible with writer type: map",
                "/",
            ),
            (
                INT_MAP_SCHEMA,
                LONG_MAP_SCHEMA,
                "reader type: int not compatible with writer type: long",
                "/values",
            ),
            (
                INT_SCHEMA,
                ENUM2_AB_SCHEMA,
                "reader type: int not compatible with writer type: enum",
                "/",
            ),
            (
                ENUM2_AB_SCHEMA,
                INT_SCHEMA,
                "reader type: enum not compatible with writer type: int",
                "/",
            ),
            (
                FLOAT_SCHEMA,
                INT_LONG_FLOAT_DOUBLE_UNION_SCHEMA,
                "reader type: float not compatible with writer type: double",
                "/",
            ),
            (
                LONG_SCHEMA,
                INT_FLOAT_UNION_SCHEMA,
                "reader type: long not compatible with writer type: float",
                "/",
            ),
            (
                INT_SCHEMA,
                INT_FLOAT_UNION_SCHEMA,
                "reader type: int not compatible with writer type: float",
                "/",
            ),
            # (INT_LIST_RECORD, LONG_LIST_RECORD, "reader type: int not compatible with writer type: long", "/fields/0/type"),
            (
                NULL_SCHEMA,
                INT_SCHEMA,
                "reader type: null not compatible with writer type: int",
                "/",
            ),
        ]
        for (reader, writer, message, location) in incompatible_pairs:
            result = ReaderWriterCompatibilityChecker().get_compatibility(reader, writer)
            self.assertIs(result.compatibility, SchemaCompatibilityType.incompatible)
            self.assertIn(message, result.messages)
            self.assertIn(location, result.locations)

    def are_compatible(self, reader: Schema, writer: Schema) -> bool:
        return ReaderWriterCompatibilityChecker().get_compatibility(reader, writer).compatibility is SchemaCompatibilityType.compatible



#!/usr/bin/env python3

##
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test the schema parsing logic."""

import json
import unittest
import warnings
from typing import List

import avro.errors
import avro.schema


class TestSchema:
    """A proxy for a schema string that provides useful test metadata."""

    def __init__(self, data, name="", comment="", warnings=None):
        if not isinstance(data, str):
            data = json.dumps(data)
        self.data = data
        self.name = name or data  # default to data for name
        self.comment = comment
        self.warnings = warnings

    def parse(self):
        return avro.schema.parse(str(self))

    def __str__(self):
        return str(self.data)


class ValidTestSchema(TestSchema):
    """A proxy for a valid schema string that provides useful test metadata."""

    valid = True


class InvalidTestSchema(TestSchema):
    """A proxy for an invalid schema string that provides useful test metadata."""

    valid = False


PRIMITIVE_EXAMPLES = [InvalidTestSchema('"True"')]  # type: List[TestSchema]
PRIMITIVE_EXAMPLES.append(InvalidTestSchema("True"))
PRIMITIVE_EXAMPLES.append(InvalidTestSchema('{"no_type": "test"}'))
PRIMITIVE_EXAMPLES.append(InvalidTestSchema('{"type": "panther"}'))
PRIMITIVE_EXAMPLES.extend([ValidTestSchema(f'"{t}"') for t in avro.schema.PRIMITIVE_TYPES])
PRIMITIVE_EXAMPLES.extend([ValidTestSchema({"type": t}) for t in avro.schema.PRIMITIVE_TYPES])

FIXED_EXAMPLES = [
    ValidTestSchema({"type": "fixed", "name": "Test", "size": 1}),
    ValidTestSchema(
        {
            "type": "fixed",
            "name": "MyFixed",
            "size": 1,
            "namespace": "org.apache.hadoop.avro",
        }
    ),
    ValidTestSchema({"type": "fixed", "name": "NullNamespace", "namespace": None, "size": 1}),
    ValidTestSchema({"type": "fixed", "name": "EmptyStringNamespace", "namespace": "", "size": 1}),
    InvalidTestSchema({"type": "fixed", "name": "Missing size"}),
    InvalidTestSchema({"type": "fixed", "size": 314}),
    InvalidTestSchema({"type": "fixed", "size": 314, "name": "dr. spaceman"}, comment="AVRO-621"),
]

ENUM_EXAMPLES = [
    ValidTestSchema({"type": "enum", "name": "Test", "symbols": ["A", "B"]}),
    ValidTestSchema({"type": "enum", "name": "AVRO2174", "symbols": ["nowhitespace"]}),
    InvalidTestSchema({"type": "enum", "name": "Status", "symbols": "Normal Caution Critical"}),
    InvalidTestSchema({"type": "enum", "name": [0, 1, 1, 2, 3, 5, 8], "symbols": ["Golden", "Mean"]}),
    InvalidTestSchema({"type": "enum", "symbols": ["I", "will", "fail", "no", "name"]}),
    InvalidTestSchema({"type": "enum", "name": "Test", "symbols": ["AA", "AA"]}),
    InvalidTestSchema({"type": "enum", "name": "AVRO2174", "symbols": ["white space"]}),
]

ARRAY_EXAMPLES = [
    ValidTestSchema({"type": "array", "items": "long"}),
    ValidTestSchema(
        {
            "type": "array",
            "items": {"type": "enum", "name": "Test", "symbols": ["A", "B"]},
        }
    ),
]

MAP_EXAMPLES = [
    ValidTestSchema({"type": "map", "values": "long"}),
    ValidTestSchema(
        {
            "type": "map",
            "values": {"type": "enum", "name": "Test", "symbols": ["A", "B"]},
        }
    ),
]

UNION_EXAMPLES = [
    ValidTestSchema(["string", "null", "long"]),
    InvalidTestSchema(["null", "null"]),
    InvalidTestSchema(["long", "long"]),
    InvalidTestSchema([{"type": "array", "items": "long"}, {"type": "array", "items": "string"}]),
]

NAMED_IN_UNION_EXAMPLES = [
    ValidTestSchema(
        {
            "namespace": "org.apache.avro.test",
            "type": "record",
            "name": "Test",
            "fields": [
                {
                    "type": {
                        "symbols": ["one", "two"],
                        "type": "enum",
                        "name": "NamedEnum",
                    },
                    "name": "thenamedenum",
                },
                {"type": ["null", "NamedEnum"], "name": "unionwithreftoenum"},
            ],
        }
    )
]

RECORD_EXAMPLES = [
    ValidTestSchema({"type": "record", "name": "Test", "fields": [{"name": "f", "type": "long"}]}),
    ValidTestSchema({"type": "error", "name": "Test", "fields": [{"name": "f", "type": "long"}]}),
    ValidTestSchema(
        {
            "type": "record",
            "name": "Node",
            "fields": [
                {"name": "label", "type": "string"},
                {"name": "children", "type": {"type": "array", "items": "Node"}},
            ],
        }
    ),
    ValidTestSchema(
        {
            "type": "record",
            "name": "Lisp",
            "fields": [
                {
                    "name": "value",
                    "type": [
                        "null",
                        "string",
                        {
                            "type": "record",
                            "name": "Cons",
                            "fields": [
                                {"name": "car", "type": "Lisp"},
                                {"name": "cdr", "type": "Lisp"},
                            ],
                        },
                    ],
                }
            ],
        }
    ),
    ValidTestSchema(
        {
            "type": "record",
            "name": "HandshakeRequest",
            "namespace": "org.apache.avro.ipc",
            "fields": [
                {
                    "name": "clientHash",
                    "type": {"type": "fixed", "name": "MD5", "size": 16},
                },
                {"name": "clientProtocol", "type": ["null", "string"]},
                {"name": "serverHash", "type": "MD5"},
                {"name": "meta", "type": ["null", {"type": "map", "values": "bytes"}]},
            ],
        }
    ),
    ValidTestSchema(
        {
            "type": "record",
            "name": "HandshakeResponse",
            "namespace": "org.apache.avro.ipc",
            "fields": [
                {
                    "name": "match",
                    "type": {
                        "type": "enum",
                        "name": "HandshakeMatch",
                        "symbols": ["BOTH", "CLIENT", "NONE"],
                    },
                },
                {"name": "serverProtocol", "type": ["null", "string"]},
                {
                    "name": "serverHash",
                    "type": ["null", {"name": "MD5", "size": 16, "type": "fixed"}],
                },
                {"name": "meta", "type": ["null", {"type": "map", "values": "bytes"}]},
            ],
        }
    ),
    ValidTestSchema(
        {
            "type": "record",
            "name": "Interop",
            "namespace": "org.apache.avro",
            "fields": [
                {"name": "intField", "type": "int"},
                {"name": "longField", "type": "long"},
                {"name": "stringField", "type": "string"},
                {"name": "boolField", "type": "boolean"},
                {"name": "floatField", "type": "float"},
                {"name": "doubleField", "type": "double"},
                {"name": "bytesField", "type": "bytes"},
                {"name": "nullField", "type": "null"},
                {"name": "arrayField", "type": {"type": "array", "items": "double"}},
                {
                    "name": "mapField",
                    "type": {
                        "type": "map",
                        "values": {
                            "name": "Foo",
                            "type": "record",
                            "fields": [{"name": "label", "type": "string"}],
                        },
                    },
                },
                {
                    "name": "unionField",
                    "type": ["boolean", "double", {"type": "array", "items": "bytes"}],
                },
                {
                    "name": "enumField",
                    "type": {
                        "type": "enum",
                        "name": "Kind",
                        "symbols": ["A", "B", "C"],
                    },
                },
                {
                    "name": "fixedField",
                    "type": {"type": "fixed", "name": "MD5", "size": 16},
                },
                {
                    "name": "recordField",
                    "type": {
                        "type": "record",
                        "name": "Node",
                        "fields": [
                            {"name": "label", "type": "string"},
                            {
                                "name": "children",
                                "type": {"type": "array", "items": "Node"},
                            },
                        ],
                    },
                },
            ],
        }
    ),
    ValidTestSchema(
        {
            "type": "record",
            "name": "ipAddr",
            "fields": [
                {
                    "name": "addr",
                    "type": [
                        {"name": "IPv6", "type": "fixed", "size": 16},
                        {"name": "IPv4", "type": "fixed", "size": 4},
                    ],
                }
            ],
        }
    ),
    InvalidTestSchema(
        {
            "type": "record",
            "name": "Address",
            "fields": [{"type": "string"}, {"type": "string", "name": "City"}],
        }
    ),
    InvalidTestSchema(
        {
            "type": "record",
            "name": "Event",
            "fields": [{"name": "Sponsor"}, {"name": "City", "type": "string"}],
        }
    ),
    InvalidTestSchema(
        {
            "type": "record",
            "name": "Rainer",
            "fields": "His vision, from the constantly passing bars",
        }
    ),
    InvalidTestSchema(
        {
            "name": ["Tom", "Jerry"],
            "type": "record",
            "fields": [{"name": "name", "type": "string"}],
        }
    ),
]

DOC_EXAMPLES = [
    ValidTestSchema(
        {
            "type": "record",
            "name": "TestDoc",
            "doc": "Doc string",
            "fields": [{"name": "name", "type": "string", "doc": "Doc String"}],
        }
    ),
    ValidTestSchema({"type": "enum", "name": "Test", "symbols": ["A", "B"], "doc": "Doc String"}),
]

OTHER_PROP_EXAMPLES = [
    ValidTestSchema(
        {
            "type": "record",
            "name": "TestRecord",
            "cp_string": "string",
            "cp_int": 1,
            "cp_array": [1, 2, 3, 4],
            "fields": [
                {"name": "f1", "type": "string", "cp_object": {"a": 1, "b": 2}},
                {"name": "f2", "type": "long", "cp_null": None},
            ],
        }
    ),
    ValidTestSchema({"type": "map", "values": "long", "cp_boolean": True}),
    ValidTestSchema(
        {
            "type": "enum",
            "name": "TestEnum",
            "symbols": ["one", "two", "three"],
            "cp_float": 1.0,
        }
    ),
]

DECIMAL_LOGICAL_TYPE = [
    ValidTestSchema(
        {
            "type": "fixed",
            "logicalType": "decimal",
            "name": "TestDecimal",
            "precision": 4,
            "size": 10,
            "scale": 2,
        }
    ),
    ValidTestSchema({"type": "bytes", "logicalType": "decimal", "precision": 4, "scale": 2}),
    InvalidTestSchema(
        {
            "type": "fixed",
            "logicalType": "decimal",
            "name": "TestDecimal2",
            "precision": 2,
            "scale": 2,
            "size": -2,
        }
    ),
]

DATE_LOGICAL_TYPE = [ValidTestSchema({"type": "int", "logicalType": "date"})]

TIMEMILLIS_LOGICAL_TYPE = [ValidTestSchema({"type": "int", "logicalType": "time-millis"})]

TIMEMICROS_LOGICAL_TYPE = [ValidTestSchema({"type": "long", "logicalType": "time-micros"})]

TIMESTAMPMILLIS_LOGICAL_TYPE = [ValidTestSchema({"type": "long", "logicalType": "timestamp-millis"})]

TIMESTAMPMICROS_LOGICAL_TYPE = [ValidTestSchema({"type": "long", "logicalType": "timestamp-micros"})]

UUID_LOGICAL_TYPE = [ValidTestSchema({"type": "string", "logicalType": "uuid"})]

IGNORED_LOGICAL_TYPE = [
    ValidTestSchema(
        {"type": "string", "logicalType": "unknown-logical-type"},
        warnings=[avro.errors.IgnoredLogicalType("Unknown unknown-logical-type, using string.")],
    ),
    ValidTestSchema(
        {"type": "bytes", "logicalType": "decimal", "scale": 0},
        warnings=[avro.errors.IgnoredLogicalType("Invalid decimal precision None. Must be a positive integer.")],
    ),
    ValidTestSchema(
        {"type": "bytes", "logicalType": "decimal", "precision": 2.4, "scale": 0},
        warnings=[avro.errors.IgnoredLogicalType("Invalid decimal precision 2.4. Must be a positive integer.")],
    ),
    ValidTestSchema(
        {"type": "bytes", "logicalType": "decimal", "precision": 2, "scale": -2},
        warnings=[avro.errors.IgnoredLogicalType("Invalid decimal scale -2. Must be a positive integer.")],
    ),
    ValidTestSchema(
        {"type": "bytes", "logicalType": "decimal", "precision": -2, "scale": 2},
        warnings=[avro.errors.IgnoredLogicalType("Invalid decimal precision -2. Must be a positive integer.")],
    ),
    ValidTestSchema(
        {"type": "bytes", "logicalType": "decimal", "precision": 2, "scale": 3},
        warnings=[avro.errors.IgnoredLogicalType("Invalid decimal scale 3. Cannot be greater than precision 2.")],
    ),
    ValidTestSchema(
        {
            "type": "fixed",
            "logicalType": "decimal",
            "name": "TestIgnored",
            "precision": -10,
            "scale": 2,
            "size": 5,
        },
        warnings=[avro.errors.IgnoredLogicalType("Invalid decimal precision -10. Must be a positive integer.")],
    ),
    ValidTestSchema(
        {
            "type": "fixed",
            "logicalType": "decimal",
            "name": "TestIgnored",
            "scale": 2,
            "size": 5,
        },
        warnings=[avro.errors.IgnoredLogicalType("Invalid decimal precision None. Must be a positive integer.")],
    ),
    ValidTestSchema(
        {
            "type": "fixed",
            "logicalType": "decimal",
            "name": "TestIgnored",
            "precision": 2,
            "scale": 3,
            "size": 2,
        },
        warnings=[avro.errors.IgnoredLogicalType("Invalid decimal scale 3. Cannot be greater than precision 2.")],
    ),
    ValidTestSchema(
        {
            "type": "fixed",
            "logicalType": "decimal",
            "name": "TestIgnored",
            "precision": 311,
            "size": 129,
        },
        warnings=[avro.errors.IgnoredLogicalType("Invalid decimal precision 311. Max is 310.")],
    ),
    ValidTestSchema(
        {"type": "float", "logicalType": "decimal", "precision": 2, "scale": 0},
        warnings=[avro.errors.IgnoredLogicalType("Logical type decimal requires literal type bytes/fixed, not float.")],
    ),
    ValidTestSchema(
        {"type": "int", "logicalType": "date1"},
        warnings=[avro.errors.IgnoredLogicalType("Unknown date1, using int.")],
    ),
    ValidTestSchema(
        {"type": "long", "logicalType": "date"},
        warnings=[avro.errors.IgnoredLogicalType("Logical type date requires literal type int, not long.")],
    ),
    ValidTestSchema(
        {"type": "int", "logicalType": "time-milis"},
        warnings=[avro.errors.IgnoredLogicalType("Unknown time-milis, using int.")],
    ),
    ValidTestSchema(
        {"type": "long", "logicalType": "time-millis"},
        warnings=[avro.errors.IgnoredLogicalType("Logical type time-millis requires literal type int, not long.")],
    ),
    ValidTestSchema(
        {"type": "long", "logicalType": "time-micro"},
        warnings=[avro.errors.IgnoredLogicalType("Unknown time-micro, using long.")],
    ),
    ValidTestSchema(
        {"type": "int", "logicalType": "time-micros"},
        warnings=[avro.errors.IgnoredLogicalType("Logical type time-micros requires literal type long, not int.")],
    ),
    ValidTestSchema(
        {"type": "long", "logicalType": "timestamp-milis"},
        warnings=[avro.errors.IgnoredLogicalType("Unknown timestamp-milis, using long.")],
    ),
    ValidTestSchema(
        {"type": "int", "logicalType": "timestamp-millis"},
        warnings=[avro.errors.IgnoredLogicalType("Logical type timestamp-millis requires literal type long, not int.")],
    ),
    ValidTestSchema(
        {"type": "long", "logicalType": "timestamp-micro"},
        warnings=[avro.errors.IgnoredLogicalType("Unknown timestamp-micro, using long.")],
    ),
    ValidTestSchema(
        {"type": "int", "logicalType": "timestamp-micros"},
        warnings=[avro.errors.IgnoredLogicalType("Logical type timestamp-micros requires literal type long, not int.")],
    ),
]

EXAMPLES = PRIMITIVE_EXAMPLES
EXAMPLES += FIXED_EXAMPLES
EXAMPLES += ENUM_EXAMPLES
EXAMPLES += ARRAY_EXAMPLES
EXAMPLES += MAP_EXAMPLES
EXAMPLES += UNION_EXAMPLES
EXAMPLES += NAMED_IN_UNION_EXAMPLES
EXAMPLES += RECORD_EXAMPLES
EXAMPLES += DOC_EXAMPLES
EXAMPLES += DECIMAL_LOGICAL_TYPE
EXAMPLES += DATE_LOGICAL_TYPE
EXAMPLES += TIMEMILLIS_LOGICAL_TYPE
EXAMPLES += TIMEMICROS_LOGICAL_TYPE
EXAMPLES += TIMESTAMPMILLIS_LOGICAL_TYPE
EXAMPLES += TIMESTAMPMICROS_LOGICAL_TYPE
EXAMPLES += UUID_LOGICAL_TYPE
EXAMPLES += IGNORED_LOGICAL_TYPE

VALID_EXAMPLES = [e for e in EXAMPLES if getattr(e, "valid", False)]
INVALID_EXAMPLES = [e for e in EXAMPLES if not getattr(e, "valid", True)]


class TestMisc(unittest.TestCase):
    """Miscellaneous tests for schema"""

    def test_correct_recursive_extraction(self):
        """A recursive reference within a schema should be the same type every time."""
        s = avro.schema.parse(
            """{
            "type": "record",
            "name": "X",
            "fields": [{
                "name": "y",
                "type": {
                    "type": "record",
                    "name": "Y",
                    "fields": [{"name": "Z", "type": "X"}]}
            }]
        }"""
        )
        t = avro.schema.parse(str(s.fields[0].type))
        # If we've made it this far, the subschema was reasonably stringified; it ccould be reparsed.
        self.assertEqual("X", t.fields[0].type.name)

    def test_name_is_none(self):
        """When a name is None its namespace is None."""
        self.assertIsNone(avro.schema.Name(None, None, None).fullname)
        self.assertIsNone(avro.schema.Name(None, None, None).space)

    def test_name_not_empty_string(self):
        """A name cannot be the empty string."""
        self.assertRaises(avro.errors.SchemaParseException, avro.schema.Name, "", None, None)

    def test_name_space_specified(self):
        """Space combines with a name to become the fullname."""
        # name and namespace specified
        fullname = avro.schema.Name("a", "o.a.h", None).fullname
        self.assertEqual(fullname, "o.a.h.a")

    def test_fullname_space_specified(self):
        """When name contains dots, namespace should be ignored."""
        fullname = avro.schema.Name("a.b.c.d", "o.a.h", None).fullname
        self.assertEqual(fullname, "a.b.c.d")

    def test_name_default_specified(self):
        """Default space becomes the namespace when the namespace is None."""
        fullname = avro.schema.Name("a", None, "b.c.d").fullname
        self.assertEqual(fullname, "b.c.d.a")

    def test_fullname_default_specified(self):
        """When a name contains dots, default space should be ignored."""
        fullname = avro.schema.Name("a.b.c.d", None, "o.a.h").fullname
        self.assertEqual(fullname, "a.b.c.d")

    def test_fullname_space_default_specified(self):
        """When a name contains dots, namespace and default space should be ignored."""
        fullname = avro.schema.Name("a.b.c.d", "o.a.a", "o.a.h").fullname
        self.assertEqual(fullname, "a.b.c.d")

    def test_name_space_default_specified(self):
        """When name and space are specified, default space should be ignored."""
        fullname = avro.schema.Name("a", "o.a.a", "o.a.h").fullname
        self.assertEqual(fullname, "o.a.a.a")

    def test_equal_names(self):
        """Equality of names is defined on the fullname and is case-sensitive."""
        self.assertEqual(
            avro.schema.Name("a.b.c.d", None, None),
            avro.schema.Name("d", "a.b.c", None),
        )
        self.assertNotEqual(avro.schema.Name("C.d", None, None), avro.schema.Name("c.d", None, None))

    def test_invalid_name(self):
        """The name portion of a fullname, record field names, and enum symbols must:
        start with [A-Za-z_] and subsequently contain only [A-Za-z0-9_]"""
        self.assertRaises(
            avro.errors.InvalidName,
            avro.schema.Name,
            "an especially spacey cowboy",
            None,
            None,
        )
        self.assertRaises(
            avro.errors.InvalidName,
            avro.schema.Name,
            "99 problems but a name aint one",
            None,
            None,
        )

    def test_null_namespace(self):
        """The empty string may be used as a namespace to indicate the null namespace."""
        name = avro.schema.Name("name", "", None)
        self.assertEqual(name.fullname, "name")
        self.assertIsNone(name.space)

    def test_exception_is_not_swallowed_on_parse_error(self):
        """A specific exception message should appear on a json parse error."""
        self.assertRaisesRegex(
            avro.errors.SchemaParseException,
            r"Error parsing JSON: /not/a/real/file",
            avro.schema.parse,
            "/not/a/real/file",
        )

    def test_decimal_valid_type(self):
        fixed_decimal_schema = ValidTestSchema(
            {
                "type": "fixed",
                "logicalType": "decimal",
                "name": "TestDecimal",
                "precision": 4,
                "scale": 2,
                "size": 2,
            }
        )

        bytes_decimal_schema = ValidTestSchema({"type": "bytes", "logicalType": "decimal", "precision": 4})

        fixed_decimal = fixed_decimal_schema.parse()
        self.assertEqual(4, fixed_decimal.get_prop("precision"))
        self.assertEqual(2, fixed_decimal.get_prop("scale"))
        self.assertEqual(2, fixed_decimal.get_prop("size"))

        bytes_decimal = bytes_decimal_schema.parse()
        self.assertEqual(4, bytes_decimal.get_prop("precision"))
        self.assertEqual(0, bytes_decimal.get_prop("scale"))

    def test_fixed_decimal_valid_max_precision(self):
        # An 8 byte number can represent any 18 digit number.
        fixed_decimal_schema = ValidTestSchema(
            {
                "type": "fixed",
                "logicalType": "decimal",
                "name": "TestDecimal",
                "precision": 18,
                "scale": 0,
                "size": 8,
            }
        )

        fixed_decimal = fixed_decimal_schema.parse()
        self.assertIsInstance(fixed_decimal, avro.schema.FixedSchema)
        self.assertIsInstance(fixed_decimal, avro.schema.DecimalLogicalSchema)

    def test_fixed_decimal_invalid_max_precision(self):
        # An 8 byte number can't represent every 19 digit number, so the logical
        # type is not applied.
        fixed_decimal_schema = ValidTestSchema(
            {
                "type": "fixed",
                "logicalType": "decimal",
                "name": "TestDecimal",
                "precision": 19,
                "scale": 0,
                "size": 8,
            }
        )

        fixed_decimal = fixed_decimal_schema.parse()
        self.assertIsInstance(fixed_decimal, avro.schema.FixedSchema)
        self.assertNotIsInstance(fixed_decimal, avro.schema.DecimalLogicalSchema)

    def test_parse_invalid_symbol(self):
        """Disabling enumschema symbol validation should allow invalid symbols to pass."""
        test_schema_string = json.dumps({"type": "enum", "name": "AVRO2174", "symbols": ["white space"]})
        with self.assertRaises(avro.errors.InvalidName, msg="When enum symbol validation is enabled, an invalid symbol should raise InvalidName."):
            avro.schema.parse(test_schema_string, validate_enum_symbols=True)
        try:
            avro.schema.parse(test_schema_string, validate_enum_symbols=False)
        except avro.errors.InvalidName:  # pragma: no coverage
            self.fail("When enum symbol validation is disabled, an invalid symbol should not raise InvalidName.")


class SchemaParseTestCase(unittest.TestCase):
    """Enable generating parse test cases over all the valid and invalid example schema."""

    def __init__(self, test_schema):
        """Ignore the normal signature for unittest.TestCase because we are generating
        many test cases from this one class. This is safe as long as the autoloader
        ignores this class. The autoloader will ignore this class as long as it has
        no methods starting with `test_`.
        """
        super().__init__("parse_valid" if test_schema.valid else "parse_invalid")
        self.test_schema = test_schema
        # Never hide repeated warnings when running this test case.
        warnings.simplefilter("always")

    def parse_valid(self) -> None:
        """Parsing a valid schema should not error, but may contain warnings."""
        test_warnings = self.test_schema.warnings or []
        try:
            warnings.filterwarnings(action="error", category=avro.errors.IgnoredLogicalType)
            self.test_schema.parse()
        except (avro.errors.IgnoredLogicalType) as e:
            self.assertIn(type(e), (type(w) for w in test_warnings))
            self.assertIn(str(e), (str(w) for w in test_warnings))
        except (avro.errors.AvroException, avro.errors.SchemaParseException):  # pragma: no coverage
            self.fail(f"Valid schema failed to parse: {self.test_schema!s}")
        else:
            self.assertEqual([], test_warnings)
        finally:
            warnings.filterwarnings(action="default", category=avro.errors.IgnoredLogicalType)

    def parse_invalid(self):
        """Parsing an invalid schema should error."""
        with self.assertRaises(
            (avro.errors.AvroException, avro.errors.SchemaParseException), msg=f"Invalid schema should not have parsed: {self.test_schema!s}"
        ):
            self.test_schema.parse()


class RoundTripParseTestCase(unittest.TestCase):
    """Enable generating round-trip parse test cases over all the valid test schema."""

    def __init__(self, test_schema):
        """Ignore the normal signature for unittest.TestCase because we are generating
        many test cases from this one class. This is safe as long as the autoloader
        ignores this class. The autoloader will ignore this class as long as it has
        no methods starting with `test_`.
        """
        super().__init__("parse_round_trip")
        self.test_schema = test_schema

    def parse_round_trip(self):
        """The string of a Schema should be parseable to the same Schema."""
        parsed = self.test_schema.parse()
        round_trip = avro.schema.parse(str(parsed))
        self.assertEqual(
            parsed,
            round_trip,
            {
                "original schema": parsed.to_json(),
                "round trip schema": round_trip.to_json(),
            },
        )


class DocAttributesTestCase(unittest.TestCase):
    """Enable generating document attribute test cases over all the document test schema."""

    def __init__(self, test_schema):
        """Ignore the normal signature for unittest.TestCase because we are generating
        many test cases from this one class. This is safe as long as the autoloader
        ignores this class. The autoloader will ignore this class as long as it has
        no methods starting with `test_`.
        """
        super().__init__("check_doc_attributes")
        self.test_schema = test_schema

    def check_doc_attributes(self):
        """Documentation attributes should be preserved."""
        sch = self.test_schema.parse()
        self.assertIsNotNone(sch.doc, f"Failed to preserve 'doc' in schema: {self.test_schema!s}")
        if sch.type == "record":
            for f in sch.fields:
                self.assertIsNotNone(
                    f.doc,
                    f"Failed to preserve 'doc' in fields: {self.test_schema!s}",
                )


class OtherAttributesTestCase(unittest.TestCase):
    """Enable generating attribute test cases over all the other-prop test schema."""

    _type_map = {
        "cp_array": list,
        "cp_boolean": bool,
        "cp_float": float,
        "cp_int": int,
        "cp_null": type(None),
        "cp_object": dict,
        "cp_string": str,
    }

    def __init__(self, test_schema):
        """Ignore the normal signature for unittest.TestCase because we are generating
        many test cases from this one class. This is safe as long as the autoloader
        ignores this class. The autoloader will ignore this class as long as it has
        no methods starting with `test_`.
        """
        super().__init__("check_attributes")
        self.test_schema = test_schema

    def _check_props(self, props):
        for k, v in props.items():
            self.assertIsInstance(v, self._type_map[k])

    def check_attributes(self):
        """Other attributes and their types on a schema should be preserved."""
        sch = self.test_schema.parse()
        try:
            self.assertNotEqual(sch, object(), "A schema is never equal to a non-schema instance.")
        except AttributeError:  # pragma: no coverage
            self.fail("Comparing a schema to a non-schema should be False, but not error.")
        round_trip = avro.schema.parse(str(sch))
        self.assertEqual(
            sch,
            round_trip,
            "A schema should be equal to another schema parsed from the same json.",
        )
        self.assertEqual(
            sch.other_props,
            round_trip.other_props,
            "Properties were not preserved in a round-trip parse.",
        )
        self._check_props(sch.other_props)
        if sch.type == "record":
            field_props = [f.other_props for f in sch.fields if f.other_props]
            self.assertEqual(len(field_props), len(sch.fields))
            for p in field_props:
                self._check_props(p)


class CanonicalFormTestCase(unittest.TestCase):
    r"""Enable generating canonical-form test cases over the valid schema.
    Transforming into Parsing Canonical Form
    Assuming an input schema (in JSON form) that's already UTF-8 text for a valid Avro schema (including all
    quotes as required by JSON), the following transformations will produce its Parsing Canonical Form:
        - [PRIMITIVES] Convert primitive schemas to their simple form (e.g., int instead of {"type":"int"}).
        - [FULLNAMES] Replace short names with fullnames, using applicable namespaces to do so. Then eliminate
            namespace attributes, which are now redundant.
        - [STRIP] Keep only attributes that are relevant to parsing data, which are: type, name, fields, symbols,
            items, values, size. Strip all others (e.g., doc and aliases).
        - [ORDER] Order the appearance of fields of JSON objects as follows: name, type, fields, symbols, items,
            values, size. For example, if an object has type, name, and size fields, then the name field should
            appear first, followed by the type and then the size fields.
        - [STRINGS] For all JSON string literals in the schema text, replace any escaped characters
            (e.g., \uXXXX escapes) with their UTF-8 equivalents.
        - [INTEGERS] Eliminate quotes around and any leading zeros in front of JSON integer literals
            (which appear in the size attributes of fixed schemas).
        - [WHITESPACE] Eliminate all whitespace in JSON outside of string literals.
    We depend on the Python json parser to properly handle the STRINGS and INTEGERS rules, so
    we don't test them here.
    """

    def compact_json_string(self, json_doc):
        """Returns compact-encoded JSON string representation for supplied document.

        Args:
            json_doc (json): JSON Document

        Returns:
            str: Compact-encoded, stringified JSON document
        """
        return json.dumps(json_doc, separators=(",", ":"))

    def test_primitive_int(self):
        """
        Convert primitive schemas to their simple form (e.g., int instead of {"type":"int"}).
        """
        s = avro.schema.parse(json.dumps("int"))
        self.assertEqual(s.canonical_form, '"int"')

        s = avro.schema.parse(json.dumps({"type": "int"}))
        self.assertEqual(s.canonical_form, '"int"')

    def test_primitive_float(self):
        s = avro.schema.parse(json.dumps("float"))
        self.assertEqual(s.canonical_form, '"float"')

        s = avro.schema.parse(json.dumps({"type": "float"}))
        self.assertEqual(s.canonical_form, '"float"')

    def test_primitive_double(self):
        s = avro.schema.parse(json.dumps("double"))
        self.assertEqual(s.canonical_form, '"double"')

        s = avro.schema.parse(json.dumps({"type": "double"}))
        self.assertEqual(s.canonical_form, '"double"')

    def test_primitive_null(self):
        s = avro.schema.parse(json.dumps("null"))
        self.assertEqual(s.canonical_form, '"null"')

        s = avro.schema.parse(json.dumps({"type": "null"}))
        self.assertEqual(s.canonical_form, '"null"')

    def test_primitive_bytes(self):
        s = avro.schema.parse(json.dumps("bytes"))
        self.assertEqual(s.canonical_form, '"bytes"')

        s = avro.schema.parse(json.dumps({"type": "bytes"}))
        self.assertEqual(s.canonical_form, '"bytes"')

    def test_primitive_long(self):
        s = avro.schema.parse(json.dumps("long"))
        self.assertEqual(s.canonical_form, '"long"')

        s = avro.schema.parse(json.dumps({"type": "long"}))
        self.assertEqual(s.canonical_form, '"long"')

    def test_primitive_boolean(self):
        s = avro.schema.parse(json.dumps("boolean"))
        self.assertEqual(s.canonical_form, '"boolean"')

        s = avro.schema.parse(json.dumps({"type": "boolean"}))
        self.assertEqual(s.canonical_form, '"boolean"')

    def test_primitive_string(self):
        s = avro.schema.parse(json.dumps("string"))
        self.assertEqual(s.canonical_form, '"string"')

        s = avro.schema.parse(json.dumps({"type": "string"}))
        self.assertEqual(s.canonical_form, '"string"')

    def test_integer_canonical_form(self):
        """
        Integer literals starting with 0 are illegal in python, because of ambiguity. This is a placeholder test
        for INTEGERS canonical form, which should generally succeed provided a valid integer has been supplied.
        """
        s = avro.schema.parse('{"name":"md5","type":"fixed","size":16}')
        self.assertEqual(
            s.canonical_form,
            self.compact_json_string({"name": "md5", "type": "fixed", "size": 16}),
        )

    def test_string_with_escaped_characters(self):
        """
        Replace any escaped characters (e.g., \u0031 escapes) with their UTF-8 equivalents.
        """
        s = avro.schema.parse('{"name":"\u0041","type":"fixed","size":16}')
        self.assertEqual(
            s.canonical_form,
            self.compact_json_string({"name": "A", "type": "fixed", "size": 16}),
        )

    def test_fullname(self):
        """
        Replace short names with fullnames, using applicable namespaces to do so. Then eliminate namespace attributes, which are now redundant.
        """
        s = avro.schema.parse(
            json.dumps(
                {
                    "namespace": "avro",
                    "name": "example",
                    "type": "enum",
                    "symbols": ["a", "b"],
                }
            )
        )
        self.assertEqual(
            s.canonical_form,
            self.compact_json_string({"name": "avro.example", "type": "enum", "symbols": ["a", "b"]}),
        )

    def test_strip(self):
        """
        Keep only attributes that are relevant to parsing data, which are: type, name, fields, symbols, items, values,
        size. Strip all others (e.g., doc and aliases).
        """
        s = avro.schema.parse(
            json.dumps(
                {
                    "name": "foo",
                    "type": "enum",
                    "doc": "test",
                    "aliases": ["bar"],
                    "symbols": ["a", "b"],
                }
            )
        )
        self.assertEqual(
            s.canonical_form,
            self.compact_json_string({"name": "foo", "type": "enum", "symbols": ["a", "b"]}),
        )

    def test_order(self):
        """
        Order the appearance of fields of JSON objects as follows: name, type, fields, symbols, items, values, size.
        For example, if an object has type, name, and size fields, then the name field should appear first, followed
        by the type and then the size fields.
        """
        s = avro.schema.parse(json.dumps({"symbols": ["a", "b"], "type": "enum", "name": "example"}))
        self.assertEqual(
            s.canonical_form,
            self.compact_json_string({"name": "example", "type": "enum", "symbols": ["a", "b"]}),
        )

    def test_whitespace(self):
        """
        Eliminate all whitespace in JSON outside of string literals.
        """
        s = avro.schema.parse(
            """{"type": "fixed",
            "size": 16,
            "name": "md5"}
                """
        )
        self.assertEqual(
            s.canonical_form,
            self.compact_json_string({"name": "md5", "type": "fixed", "size": 16}),
        )

    def test_record_field(self):
        """
        Ensure that record fields produce the correct parsing canonical form.
        """
        s = avro.schema.parse(
            json.dumps(
                {
                    "type": "record",
                    "name": "Test",
                    "doc": "This is a test schema",
                    "aliases": ["also", "known", "as"],
                    "fields": [
                        {
                            "type": {
                                "symbols": ["one", "two"],
                                "type": "enum",
                                "name": "NamedEnum",
                            },
                            "name": "thenamedenum",
                            "doc": "This is a named enum",
                        },
                        {"type": ["null", "NamedEnum"], "name": "unionwithreftoenum"},
                    ],
                }
            )
        )
        expected = self.compact_json_string(
            {
                "name": "Test",
                "type": "record",
                "fields": [
                    {
                        "name": "thenamedenum",
                        "type": {
                            "name": "NamedEnum",
                            "type": "enum",
                            "symbols": ["one", "two"],
                        },
                    },
                    {"name": "unionwithreftoenum", "type": ["null", "NamedEnum"]},
                ],
            }
        )
        self.assertEqual(s.canonical_form, expected)

    def test_array(self):
        """
        Ensure that array schema produce the correct parsing canonical form.
        """
        s = avro.schema.parse(json.dumps({"items": "long", "type": "array"}))
        self.assertEqual(
            s.canonical_form,
            self.compact_json_string({"type": "array", "items": "long"}),
        )

    def test_map(self):
        """
        Ensure that map schema produce the correct parsing canonical form.
        """
        s = avro.schema.parse(json.dumps({"values": "long", "type": "map"}))
        self.assertEqual(
            s.canonical_form,
            self.compact_json_string({"type": "map", "values": "long"}),
        )

    def test_union(self):
        """
        Ensure that a union schema produces the correct parsing canonical form.
        """
        s = avro.schema.parse(json.dumps(["string", "null", "long"]))
        self.assertEqual(s.canonical_form, '["string","null","long"]')

    def test_large_record_handshake_request(self):
        s = avro.schema.parse(
            """
            {
            "type": "record",
            "name": "HandshakeRequest",
            "namespace": "org.apache.avro.ipc",
            "fields": [
                {
                "name": "clientHash",
                "type": {"type": "fixed", "name": "MD5", "size": 16}
                },
                {"name": "clientProtocol", "type": ["null", "string"]},
                {"name": "serverHash", "type": "MD5"},
                {
                "name": "meta",
                "type": ["null", {"type": "map", "values": "bytes"}]
                }
            ]
            }
            """
        )
        self.assertEqual(
            s.canonical_form,
            (
                '{"name":"org.apache.avro.ipc.HandshakeRequest","type":"record",'
                '"fields":[{"name":"clientHash","type":{"name":"org.apache.avro.ipc.MD5",'
                '"type":"fixed","size":16}},{"name":"clientProtocol","type":["null","string"]},'
                '{"name":"serverHash","type":{"name":"org.apache.avro.ipc.MD5","type":"fixed","size":16}},'
                '{"name":"meta","type":["null",{"type":"map","values":"bytes"}]}]}'
            ),
        )

    def test_large_record_handshake_response(self):
        s = avro.schema.parse(
            """
            {
            "type": "record",
            "name": "HandshakeResponse",
            "namespace": "org.apache.avro.ipc",
            "fields": [
                {
                "name": "match",
                "type": {
                    "type": "enum",
                    "name": "HandshakeMatch",
                    "symbols": ["BOTH", "CLIENT", "NONE"]
                }
                },
                {"name": "serverProtocol", "type": ["null", "string"]},
                {
                "name": "serverHash",
                "type": ["null", {"name": "MD5", "size": 16, "type": "fixed"}]
                },
                {
                "name": "meta",
                "type": ["null", {"type": "map", "values": "bytes"}]}]
                }
            """
        )
        self.assertEqual(
            s.canonical_form,
            (
                '{"name":"org.apache.avro.ipc.HandshakeResponse","type":"rec'
                'ord","fields":[{"name":"match","type":{"name":"org.apache.a'
                'vro.ipc.HandshakeMatch","type":"enum","symbols":["BOTH","CL'
                'IENT","NONE"]}},{"name":"serverProtocol","type":["null","st'
                'ring"]},{"name":"serverHash","type":["null",{"name":"org.ap'
                'ache.avro.ipc.MD5","type":"fixed","size":16}]},{"name":"met'
                'a","type":["null",{"type":"map","values":"bytes"}]}]}'
            ),
        )

    def test_large_record_interop(self):
        s = avro.schema.parse(
            """
            {
            "type": "record",
            "name": "Interop",
            "namespace": "org.apache.avro",
            "fields": [
                {"name": "intField", "type": "int"},
                {"name": "longField", "type": "long"},
                {"name": "stringField", "type": "string"},
                {"name": "boolField", "type": "boolean"},
                {"name": "floatField", "type": "float"},
                {"name": "doubleField", "type": "double"},
                {"name": "bytesField", "type": "bytes"},
                {"name": "nullField", "type": "null"},
                {"name": "arrayField", "type": {"type": "array", "items": "double"}},
                {
                "name": "mapField",
                "type": {
                    "type": "map",
                    "values": {"name": "Foo",
                            "type": "record",
                            "fields": [{"name": "label", "type": "string"}]}
                }
                },
                {
                "name": "unionField",
                "type": ["boolean", "double", {"type": "array", "items": "bytes"}]
                },
                {
                "name": "enumField",
                "type": {"type": "enum", "name": "Kind", "symbols": ["A", "B", "C"]}
                },
                {
                "name": "fixedField",
                "type": {"type": "fixed", "name": "MD5", "size": 16}
                },
                {
                "name": "recordField",
                "type": {"type": "record",
                        "name": "Node",
                        "fields": [{"name": "label", "type": "string"},
                                    {"name": "children",
                                    "type": {"type": "array",
                                                "items": "Node"}}]}
                }
            ]
            }
            """
        )
        self.assertEqual(
            s.canonical_form,
            (
                '{"name":"org.apache.avro.Interop","type":"record","fields":[{"na'
                'me":"intField","type":"int"},{"name":"longField","type":"long"},'
                '{"name":"stringField","type":"string"},{"name":"boolField","type'
                '":"boolean"},{"name":"floatField","type":"float"},{"name":"doubl'
                'eField","type":"double"},{"name":"bytesField","type":"bytes"},{"'
                'name":"nullField","type":"null"},{"name":"arrayField","type":{"t'
                'ype":"array","items":"double"}},{"name":"mapField","type":{"type'
                '":"map","values":{"name":"org.apache.avro.Foo","type":"record","'
                'fields":[{"name":"label","type":"string"}]}}},{"name":"unionFiel'
                'd","type":["boolean","double",{"type":"array","items":"bytes"}]}'
                ',{"name":"enumField","type":{"name":"org.apache.avro.Kind","type'
                '":"enum","symbols":["A","B","C"]}},{"name":"fixedField","type":{'
                '"name":"org.apache.avro.MD5","type":"fixed","size":16}},{"name":'
                '"recordField","type":{"name":"org.apache.avro.Node","type":"reco'
                'rd","fields":[{"name":"label","type":"string"},{"name":"children'
                '","type":{"type":"array","items":"org.apache.avro.Node"}}]}}]}'
            ),
        )


def load_tests(loader, default_tests, pattern):
    """Generate test cases across many test schema."""
    suite = unittest.TestSuite()
    suite.addTests(loader.loadTestsFromTestCase(TestMisc))
    suite.addTests(SchemaParseTestCase(ex) for ex in EXAMPLES)
    suite.addTests(RoundTripParseTestCase(ex) for ex in VALID_EXAMPLES)
    suite.addTests(DocAttributesTestCase(ex) for ex in DOC_EXAMPLES)
    suite.addTests(OtherAttributesTestCase(ex) for ex in OTHER_PROP_EXAMPLES)
    suite.addTests(loader.loadTestsFromTestCase(CanonicalFormTestCase))
    return suite


if __name__ == "__main__":  # pragma: no coverage
    unittest.main()



#!/usr/bin/env python3

##
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json


def _safe_pretty(schema):
    """Try to pretty-print a schema, but never raise an exception within another exception."""
    try:
        return json.dumps(json.loads(str(schema)), indent=2)
    except Exception:  # Never raise an exception within another exception.
        return schema


class AvroException(Exception):
    """The base class for exceptions in avro."""


class SchemaParseException(AvroException):
    """Raised when a schema failed to parse."""


class InvalidName(SchemaParseException):
    """User attempted to parse a schema with an invalid name."""


class AvroWarning(UserWarning):
    """Base class for warnings."""


class IgnoredLogicalType(AvroWarning):
    """Warnings for unknown or invalid logical types."""


class AvroTypeException(AvroException):
    """Raised when datum is not an example of schema."""

    def __init__(self, *args):
        try:
            expected_schema, name, datum = args[:3]
        except (IndexError, ValueError):
            return super().__init__(*args)
        pretty_expected = json.dumps(json.loads(str(expected_schema)), indent=2)
        return super().__init__(f'The datum "{datum}" provided for "{name}" is not an example of the schema {pretty_expected}')


class InvalidDefaultException(AvroTypeException):
    """Raised when a default value isn't a suitable type for the schema."""


class AvroOutOfScaleException(AvroTypeException):
    """Raised when attempting to write a decimal datum with an exponent too large for the decimal schema."""

    def __init__(self, *args):
        try:
            scale, datum, exponent = args[:3]
        except (IndexError, ValueError):
            return super().__init__(*args)
        return super().__init__(f"The exponent of {datum}, {exponent}, is too large for the schema scale of {scale}")


class SchemaResolutionException(AvroException):
    def __init__(self, fail_msg, writers_schema=None, readers_schema=None, *args):
        writers_message = f"\nWriter's Schema: {_safe_pretty(writers_schema)}" if writers_schema else ""
        readers_message = f"\nReader's Schema: {_safe_pretty(readers_schema)}" if readers_schema else ""
        super().__init__((fail_msg or "") + writers_message + readers_message, *args)


class DataFileException(AvroException):
    """Raised when there's a problem reading or writing file object containers."""


class IONotReadyException(AvroException):
    """Raised when attempting an avro operation on an io object that isn't fully initialized."""


class AvroRemoteException(AvroException):
    """Raised when an error message is sent by an Avro requestor or responder."""


class ConnectionClosedException(AvroException):
    """Raised when attempting IPC on a closed connection."""


class ProtocolParseException(AvroException):
    """Raised when a protocol failed to parse."""


class UnsupportedCodec(NotImplementedError, AvroException):
    """Raised when the compression named cannot be used."""


class UsageError(RuntimeError, AvroException):
    """An exception raised when incorrect arguments were passed."""


class AvroRuntimeException(RuntimeError, AvroException):
    """Raised when compatibility parsing encounters an unknown type"""



#!/usr/bin/env python3

##
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Contains the Name classes."""
from typing import TYPE_CHECKING, Dict, Optional

from avro.constants import VALID_TYPES

if TYPE_CHECKING:
    from avro.schema import NamedSchema

import re
import warnings

import avro.errors

# The name portion of a fullname, record field names, and enum symbols must:
# start with [A-Za-z_]
# subsequently contain only [A-Za-z0-9_]
_BASE_NAME_PATTERN = re.compile(r"(?:^|\.)[A-Za-z_][A-Za-z0-9_]*$")


def validate_basename(basename: str) -> None:
    """Raise InvalidName if the given basename is not a valid name."""
    if not _BASE_NAME_PATTERN.search(basename):
        raise avro.errors.InvalidName(f"{basename!s} is not a valid Avro name because it does not match the pattern {_BASE_NAME_PATTERN.pattern!s}")


def _validate_fullname(fullname: str) -> None:
    for name in fullname.split("."):
        validate_basename(name)


class Name:
    """Class to describe Avro name."""

    _full: Optional[str] = None

    def __init__(self, name_attr: Optional[str] = None, space_attr: Optional[str] = None, default_space: Optional[str] = None) -> None:
        """The fullname is determined in one of the following ways:

        - A name and namespace are both specified. For example, one might use "name": "X",
            "namespace": "org.foo" to indicate the fullname org.foo.X.
        - A fullname is specified. If the name specified contains a dot,
            then it is assumed to be a fullname, and any namespace also specified is ignored.
            For example, use "name": "org.foo.X" to indicate the fullname org.foo.X.
        - A name only is specified, i.e., a name that contains no dots.
            In this case the namespace is taken from the most tightly enclosing schema or protocol.
            For example, if "name": "X" is specified, and this occurs within a field of
            the record definition of org.foo.Y, then the fullname is org.foo.X.
            If there is no enclosing namespace then the null namespace is used.

        References to previously defined names are as in the latter two cases above:
        if they contain a dot they are a fullname,
        if they do not contain a dot, the namespace is the namespace of the enclosing definition.

        @arg name_attr: name value read in schema or None.
        @arg space_attr: namespace value read in schema or None. The empty string may be used as a namespace
            to indicate the null namespace.
        @arg default_space: the current default space or None.
        """
        if name_attr is None:
            return
        if name_attr == "":
            raise avro.errors.SchemaParseException("Name must not be the empty string.")
        # The empty string may be used as a namespace to indicate the null namespace.
        self._full = (
            name_attr
            if "." in name_attr or space_attr == "" or not (space_attr or default_space)
            else f"{space_attr or default_space!s}.{name_attr!s}"
        )
        _validate_fullname(self._full)

    def __eq__(self, other: object) -> bool:
        """Equality of names is defined on the fullname and is case-sensitive."""
        return hasattr(other, "fullname") and self.fullname == getattr(other, "fullname")

    @property
    def fullname(self) -> Optional[str]:
        return self._full

    @property
    def space(self) -> Optional[str]:
        """Back out a namespace from full name."""
        full = self._full or ""
        return full.rsplit(".", 1)[0] if "." in full else None

    def get_space(self) -> Optional[str]:
        warnings.warn("Name.get_space() is deprecated in favor of Name.space")
        return self.space


class Names:
    """Track name set and default namespace during parsing."""

    names: Dict[str, "NamedSchema"]

    def __init__(self, default_namespace: Optional[str] = None) -> None:
        self.names = {}
        self.default_namespace = default_namespace

    def has_name(self, name_attr: str, space_attr: Optional[str] = None) -> bool:
        test = Name(name_attr, space_attr, self.default_namespace).fullname
        return test in self.names

    def get_name(self, name_attr: str, space_attr: Optional[str] = None) -> Optional["NamedSchema"]:
        test = Name(name_attr, space_attr, self.default_namespace).fullname
        return None if test is None else self.names.get(test)

    def prune_namespace(self, properties: Dict[str, object]) -> Dict[str, object]:
        """given a properties, return properties with namespace removed if
        it matches the own default namespace"""
        if self.default_namespace is None:
            # I have no default -- no change
            return properties

        if "namespace" not in properties:
            # he has no namespace - no change
            return properties

        if properties["namespace"] != self.default_namespace:
            # we're different - leave his stuff alone
            return properties

        # we each have a namespace and it's redundant. delete his.
        prunable = properties.copy()
        del prunable["namespace"]
        return prunable

    def add_name(self, name_attr: str, space_attr: Optional[str], new_schema: "NamedSchema") -> Name:
        """
        Add a new schema object to the name set.

        @arg name_attr: name value read in schema
        @arg space_attr: namespace value read in schema.

        @return: the Name that was just added.
        """
        to_add = Name(name_attr, space_attr, self.default_namespace)

        if to_add.fullname in VALID_TYPES:
            raise avro.errors.SchemaParseException(f"{to_add.fullname} is a reserved type name.")
        if to_add.fullname in self.names:
            raise avro.errors.SchemaParseException(f'The name "{to_add.fullname}" is already in use.')
        if to_add.fullname is None:
            raise avro.errors.SchemaParseException(f'The name built from "{space_attr}.{name_attr}" is None')

        self.names[to_add.fullname] = new_schema
        return to_add



#!/usr/bin/env python3

##
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import distutils.errors
import glob
import os
import subprocess

import setuptools  # type: ignore

_HERE = os.path.dirname(os.path.abspath(__file__))
_AVRO_DIR = os.path.join(_HERE, "avro")
_VERSION_FILE_NAME = "VERSION.txt"


def _is_distribution():
    """Tests whether setup.py is invoked from a distribution.

    Returns:
        True if setup.py runs from a distribution.
        False otherwise, ie. if setup.py runs from a version control work tree.
    """
    # If a file PKG-INFO exists as a sibling of setup.py,
    # assume we are running as source distribution:
    return os.path.exists(os.path.join(_HERE, "PKG-INFO"))


def _generate_package_data():
    """Generate package data.

    This data will already exist in a distribution package,
    so this function only runs for local version control work tree.
    """
    distutils.log.info("Generating package data")

    # Avro top-level source directory:
    root_dir = os.path.dirname(os.path.dirname(_HERE))
    share_dir = os.path.join(root_dir, "share")

    # Create a PEP440 compliant version file.
    version_file_path = os.path.join(share_dir, _VERSION_FILE_NAME)
    with open(version_file_path, "rb") as vin:
        version = vin.read().replace(b"-", b"+")
    with open(os.path.join(_AVRO_DIR, _VERSION_FILE_NAME), "wb") as vout:
        vout.write(version)

    avro_schemas_dir = os.path.join(share_dir, "schemas", "org", "apache", "avro")
    ipc_dir = os.path.join(avro_schemas_dir, "ipc")
    tether_dir = os.path.join(avro_schemas_dir, "mapred", "tether")

    # Copy necessary avsc files:
    avsc_files = (
        ((share_dir, "test", "schemas", "interop.avsc"), ("",)),
        ((ipc_dir, "HandshakeRequest.avsc"), ("",)),
        ((ipc_dir, "HandshakeResponse.avsc"), ("",)),
        ((tether_dir, "InputProtocol.avpr"), ("tether",)),
        ((tether_dir, "OutputProtocol.avpr"), ("tether",)),
    )

    for src, dst in avsc_files:
        src = os.path.join(*src)
        dst = os.path.join(_AVRO_DIR, *dst)
        distutils.file_util.copy_file(src, dst)


class GenerateInteropDataCommand(setuptools.Command):
    """A command to generate Avro files for data interop test."""

    user_options = [
        ("schema-file=", None, "path to input Avro schema file"),
        ("output-path=", None, "path to output Avro data files"),
    ]

    def initialize_options(self):
        self.schema_file = os.path.join(_AVRO_DIR, "interop.avsc")
        self.output_path = os.path.join(_AVRO_DIR, "test", "interop", "data")

    def finalize_options(self):
        pass

    def run(self):
        # Late import -- this can only be run when avro is on the pythonpath,
        # more or less after install.
        import avro.test.gen_interop_data

        if not os.path.exists(self.output_path):
            os.makedirs(self.output_path)
        with open(self.schema_file) as schema_file, open(os.path.join(self.output_path, "py.avro"), "wb") as output:
            avro.test.gen_interop_data.generate(schema_file, output)


def _get_version():
    curdir = os.getcwd()
    if os.path.isfile("avro/VERSION.txt"):
        version_file = "avro/VERSION.txt"
    else:
        index = curdir.index("lang/py")
        path = curdir[:index]
        version_file = os.path.join(path, "share/VERSION.txt")
    with open(version_file) as verfile:
        # To follow the naming convention defined by PEP 440
        # in the case that the version is like "x.y.z-SNAPSHOT"
        return verfile.read().rstrip().replace("-", "+")


def main():
    if not _is_distribution():
        _generate_package_data()

    setuptools.setup(
        cmdclass={
            "generate_interop_data": GenerateInteropDataCommand,
        }
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

##
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import http.server
import json
from typing import Mapping

import avro.ipc
import avro.protocol

MAIL_PROTOCOL_JSON = json.dumps(
    {
        "namespace": "example.proto",
        "protocol": "Mail",
        "types": [
            {
                "name": "Message",
                "type": "record",
                "fields": [{"name": "to", "type": "string"}, {"name": "from", "type": "string"}, {"name": "body", "type": "string"}],
            }
        ],
        "messages": {
            "send": {"request": [{"name": "message", "type": "Message"}], "response": "string"},
            "replay": {"request": [], "response": "string"},
        },
    }
)
MAIL_PROTOCOL = avro.protocol.parse(MAIL_PROTOCOL_JSON)
SERVER_ADDRESS = ("localhost", 9090)


class MailResponder(avro.ipc.Responder):
    def __init__(self) -> None:
        super().__init__(MAIL_PROTOCOL)

    def invoke(self, message: avro.protocol.Message, request: Mapping[str, Mapping[str, str]]) -> str:
        if message.name == "send":
            return f"Sent message to {request['message']['to']} from {request['message']['from']} with body {request['message']['body']}"
        if message.name == "replay":
            return "replay"
        raise RuntimeError


class MailHandler(http.server.BaseHTTPRequestHandler):
    def do_POST(self) -> None:
        self.responder = MailResponder()
        call_request_reader = avro.ipc.FramedReader(self.rfile)
        call_request = call_request_reader.read_framed_message()
        resp_body = self.responder.respond(call_request)
        self.send_response(200)
        self.send_header("Content-Type", "avro/binary")
        self.end_headers()
        resp_writer = avro.ipc.FramedWriter(self.wfile)
        resp_writer.write_framed_message(resp_body)


def main():
    mail_server = http_server.HTTPServer(SERVER_ADDRESS, MailHandler)
    mail_server.allow_reuse_address = True
    mail_server.serve_forever()


if __name__ == "__main__":
    main()
